# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Private module with AST parser for safe evaluation."""
import ast
from typing import Any, Callable, Dict, List

import yaml


def _process_node(node: Any, flat_dict: dict) -> Any:
    """Compute an AST from the root by replacing param name by their values.

    The AST can contain can contain any parameter name of the configuration.
    The most usefull operators and built-in functions are supported,
    the random and math packages are also supported as well as some (safe)
    numpy, jax, tensorflow, pytorch functions. If/else statements and
    comprehension lists are also supported.
    """
    # Case None, bool or number
    if isinstance(node, ast.Constant):
        return node.n
    functions: Dict[Any, Callable[[Any, dict], Any]] = {
        ast.BinOp: _process_binop,  # binary operation
        ast.BoolOp: _process_boolop,  # boolean operation
        ast.Compare: _process_comparator,  # comparison
        ast.Name: _process_param_name,  # parameter name
        ast.Attribute: _process_subconfig,  # sub-config
        ast.IfExp: _process_ifexp,  # if/else
        ast.List: _process_ltd,  # list
        ast.Tuple: _process_ltd,  # tuple
        ast.Dict: _process_ltd,  # dict
        ast.Call: _process_call,  # function
        ast.ListComp: _process_listcomp,  # comprehension list
        ast.comprehension: _process_comprehension,  # comprehension
    }
    if isinstance(node, tuple(functions.keys())):
        return functions[type(node)](node, flat_dict)
    raise ValueError(f"Not supported node in expression (of type {type(node)}).")


def _process_binop(node: Any, flat_dict: dict) -> Any:
    """Process a binary operator node."""
    left_val = _process_node(node=node.left, flat_dict=flat_dict)
    right_val = _process_node(node=node.right, flat_dict=flat_dict)
    operators: Dict[Any, Callable[[Any, Any], Any]] = {
        ast.Add: lambda x, y: x + y,
        ast.Sub: lambda x, y: x - y,
        ast.Mult: lambda x, y: x * y,
        ast.Div: lambda x, y: x / y,
        ast.Pow: lambda x, y: x**y,
        ast.FloorDiv: lambda x, y: x // y,
        ast.Mod: lambda x, y: x % y,
        ast.BitOr: lambda x, y: x or y,
        ast.BitAnd: lambda x, y: x and y,
        ast.MatMult: lambda x, y: x @ y,
    }
    if isinstance(node.op, tuple(operators.keys())):
        return operators[type(node.op)](left_val, right_val)
    raise ValueError(
        f"Invalid operator detected: {node.op}."
        "Please use only these ops: '+', '-', '*', '/', '**', "
        "'//', '%', '|', '&'."
    )


def _process_boolop(node: Any, flat_dict: dict) -> Any:
    """Process bool operator node."""
    values = [
        _process_node(node=nodeval, flat_dict=flat_dict) for nodeval in node.values
    ]
    operators: Dict[Any, Callable[[Any, Any], Any]] = {
        ast.And: lambda x, y: x and y,
        ast.Or: lambda x, y: x or y,
    }
    operator = operators[type(node.op)]
    result = isinstance(node.op, ast.And)  # Neutral for the operation
    for val in values:
        result = operator(result, val)  # type: ignore
    return result


def _process_comparator(node: Any, flat_dict: dict) -> Any:
    """Process a comparator node."""
    left_val = _process_node(node=node.left, flat_dict=flat_dict)
    right_val = _process_node(node=node.comparators[0], flat_dict=flat_dict)
    operators: Dict[Any, Callable[[bool, bool], bool]] = {
        ast.Eq: lambda x, y: x == y,
        ast.NotEq: lambda x, y: x != y,
        ast.Lt: lambda x, y: x < y,
        ast.LtE: lambda x, y: x <= y,
        ast.Gt: lambda x, y: x > y,
        ast.GtE: lambda x, y: x >= y,
    }
    return operators[type(node.ops[0])](left_val, right_val)  # type: ignore


def _process_param_name(node: Any, flat_dict: dict) -> Any:
    """Process a parameter name."""
    name = node.id
    if name in flat_dict:
        if isinstance(flat_dict[name], (bool, int, float, complex, list, type(None))):
            return flat_dict[name]
        raise ValueError(
            f"Invalid value in expression for parameter '{name}', "
            f"found type {type(flat_dict[name])}, expected, None, bool, int, "
            "float, complex or list."
        )
    raise ValueError(f"Unknown parameter '{name}'.")


def _process_subconfig(node: Any, flat_dict: dict) -> Any:
    """Process a sub-config."""
    # Get recursively the attribute names
    param_name = ""
    while isinstance(node, ast.Attribute):
        param_name = f"{node.attr}.{param_name}"
        node = node.value
    # Add the global subconfig name and process the node containing the
    # full flat parameter name
    param_name = f"{node.id}.{param_name}"[:-1]  # Remove last dot
    node = ast.Name(id=param_name, ctx=ast.Load())
    return _process_node(node=node, flat_dict=flat_dict)


def _process_ifexp(node: Any, flat_dict: dict) -> Any:
    """Process a if/exp statement."""
    test_val = _process_node(node=node.test, flat_dict=flat_dict)
    return (
        _process_node(node=node.body, flat_dict=flat_dict)
        if test_val
        else _process_node(node=node.orelse, flat_dict=flat_dict)
    )


def _process_ltd(node: Any, flat_dict: dict) -> Any:
    """Process a list, a tuple or a dict node."""
    if isinstance(node, ast.List):
        # List
        return [
            _process_node(node=element, flat_dict=flat_dict) for element in node.elts
        ]
    if isinstance(node, ast.Tuple):
        # Tuple
        return tuple(
            _process_node(node=element, flat_dict=flat_dict) for element in node.elts
        )
    # Dict
    return {
        _process_node(node=key, flat_dict=flat_dict): _process_node(
            node=value, flat_dict=flat_dict
        )
        for (key, value) in zip(node.keys, node.values)
    }


def _process_call(node: Any, flat_dict: dict) -> Any:
    """Process a function node."""
    functions: Dict[str, Callable] = {
        "len": len,
        "sum": sum,
        "max": max,
        "min": min,
        "abs": abs,
        "round": round,
        "all": all,
        "any": any,
        "range": range,
        "bool": bool,
        "int": int,
        "float": float,
        "str": str,
        "list": list,
        "tuple": tuple,
        "dict": dict,
        "set": set,
    }
    args = [_process_node(node=arg, flat_dict=flat_dict) for arg in node.args]
    kwargs = {
        kwarg.arg: _process_node(node=kwarg.value, flat_dict=flat_dict)
        for kwarg in node.keywords
    }
    if isinstance(node.func, ast.Name):
        func_name = node.func.id
        value = functions[func_name](*args, **kwargs)
        return value

    # ast.Attribute
    func = _find_function(node=node.func, flat_dict=flat_dict)
    return func(*args, **kwargs)


def _find_function(node: Any, flat_dict: dict) -> Callable:
    """Find a function from node."""
    list_names: List[str] = []
    while isinstance(node, ast.Attribute):
        list_names = [str(node.attr)] + list_names
        node = node.value
    list_names = [node.id] + list_names

    aliases = {"np": "numpy", "tf": "tensorflow"}
    if list_names[0] in aliases:
        list_names[0] = aliases[list_names[0]]
    list_names = (
        ["jax", "numpy"] + list_names[1:] if list_names[0] == "jnp" else list_names
    )

    if list_names[0] in flat_dict:
        obj = flat_dict[list_names[0]]
    elif _filter_allowed(list_names=list_names):
        obj = __import__(list_names[0])
    else:
        raise ValueError(
            f"Package or function not allowed or not supported: {'.'.join(list_names)}"
        )
    for name in list_names[1:-1]:
        obj = getattr(obj, name)
    func = getattr(obj, list_names[-1])
    return func


def _filter_allowed(list_names: List[str]) -> bool:
    """Filter the allowed functions."""
    with open(
        "cliconfig/processing/allowed_functions.yaml", encoding="utf-8"
    ) as yaml_file:
        allowed_funcs = yaml.safe_load(yaml_file)
    list_names = list_names.copy()
    # jax and numpy share the same allowed functions, jax.random is also allowed
    list_names = list_names[1:] if list_names[0] == "jax" else list_names

    if any(name.startswith("_") for name in list_names):
        return False
    if list_names[0] in ("random", "math"):
        return True
    if list_names[0] in ("numpy", "torch", "tensorflow"):
        return list_names[1] in allowed_funcs[list_names[0]]
    return False


def _process_listcomp(node: Any, flat_dict: dict) -> Any:
    """Process comprehension list node."""
    elt = node.elt
    generator = node.generators[0]
    result = []
    for variables in _process_node(generator, flat_dict=flat_dict):
        result.append(_process_node(elt, flat_dict=variables))
    return result


def _process_comprehension(node: Any, flat_dict: dict) -> Any:
    """Process comprehension node."""
    target = node.target
    iterator = node.iter
    tests = node.ifs
    if isinstance(target, ast.Tuple):
        target = tuple(elt for elt in target.elts)

    variables = flat_dict.copy()
    for val in _process_node(node=iterator, flat_dict=flat_dict):
        if isinstance(target, tuple):
            variables.update({target.id: val[i] for i, target in enumerate(target)})
        else:
            variables[target.id] = val
        condition = all(_process_node(test, variables) for test in tests)
        if condition:
            yield variables



# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Module to convert yaml tags in yaml file to python dict with  cliconfig tags."""

from typing import Any, Dict, Optional, Tuple

import yaml


class TaggedNode:
    """Node class for tagged tree (with yaml tags)."""

    def __init__(self, value: Any, tag: str, *, is_config: bool) -> None:
        self.tag = tag
        self.value = value
        self.is_config = is_config


def tagged_constructor(
    loader: yaml.SafeLoader, tag_suffix: str, node: yaml.Node
) -> Any:
    """Build a tagged tree node from yaml node."""
    if isinstance(node, yaml.ScalarNode):
        return TaggedNode(loader.construct_scalar(node), tag_suffix, is_config=False)
    if isinstance(node, yaml.SequenceNode):
        return TaggedNode(loader.construct_sequence(node), tag_suffix, is_config=False)
    return TaggedNode(
        loader.construct_mapping(node), tag_suffix, is_config=True  # type: ignore
    )


def get_yaml_loader() -> Any:
    """Return a yaml loader to parse tags and build tagged tree."""
    loader = yaml.SafeLoader
    loader.add_multi_constructor("", tagged_constructor)  # type: ignore
    return loader


def insert_tags(tagged_tree: Any) -> Tuple[Any, Optional[str]]:
    """Build dict with cliconfig tag from tagged tree (with yaml tags)."""
    if isinstance(tagged_tree, TaggedNode) and tagged_tree.is_config:
        tag = tagged_tree.tag
        if tag:
            tag = tag.replace("!", "")
        out_dict = build_tree_from_dict(tagged_tree.value)
        return out_dict, tag
    if isinstance(tagged_tree, dict):
        out_dict = build_tree_from_dict(tagged_tree)
        return out_dict, None
    if isinstance(tagged_tree, TaggedNode):
        tag = tagged_tree.tag
        if tag:
            tag = tag.replace("!", "")
        if isinstance(tagged_tree.value, str):
            value = yaml.safe_load(tagged_tree.value)
        else:
            value = tagged_tree.value
        return value, tag
    return tagged_tree, None


def build_tree_from_dict(in_dict: Dict) -> Dict[str, Any]:
    """Build a dict with cliconfig tag from a dict of tagged trees."""
    out_dict: Dict[str, Any] = {}
    for key, value in in_dict.items():
        tree, tag = insert_tags(value)
        if tag:
            subtag = tag.replace("!", "")
            key = f"{key}@{subtag}"
        out_dict[key] = tree
    return out_dict



# Copyright © 2023  Valentin Goldité
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the MIT License.
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#    This project is free to use for COMMERCIAL USE, MODIFICATION,
#    DISTRIBUTION and PRIVATE USE as long as the original license is
#    include as well as this copy right notice.
"""CLI Config: build your configuration from CLI by merging with processing."""
from cliconfig._version import __version__, __version_tuple__
from cliconfig.base import Config
from cliconfig.config_routines import (
    flatten_config,
    load_config,
    make_config,
    save_config,
    show_config,
    unflatten_config,
    update_config,
)
from cliconfig.dict_routines import flatten as flatten_dict
from cliconfig.dict_routines import unflatten as unflatten_dict
from cliconfig.process_routines import (
    merge_flat_paths_processing,
    merge_flat_processing,
)
from cliconfig.processing.base import Processing
from cliconfig.processing.builtin import DefaultProcessings
from cliconfig.processing.create import (
    create_processing_keep_property,
    create_processing_value,
)

__all__ = [
    "__version__",
    "__version_tuple__",
    "Config",
    "DefaultProcessings",
    "Processing",
    "create_processing_keep_property",
    "create_processing_value",
    "flatten_config",
    "flatten_dict",
    "make_config",
    "load_config",
    "merge_flat_paths_processing",
    "merge_flat_processing",
    "save_config",
    "show_config",
    "unflatten_config",
    "unflatten_dict",
    "update_config",
]



# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Private module with type parser for processing module with type manipulation."""
from pydoc import locate
from typing import List, Optional, Tuple, Type, Union


def _parse_type(type_desc: str) -> Tuple:
    """Parse a type description.

    Allow basic types (none, any, bool, int, float, str, list, dict), nested lists,
    nested dicts, unions (with Union or the '|' symbol) and Optional.

    Examples of representation:
     * "str" -> (str,)
     * "None|Any" -> (type(None), object)
     * "list|float" -> (list, float)
     * "List[float]" -> (("list", (float,)),)
     * "Dict[str, Any]" -> (("dict", (str,), (object,)),)
     * "Dict[str, List[float]]" -> (("dict", (str,), ((("list", (float,)),),)),)
     * "Dict[str|bool, int|float]" -> (("dict", (str, bool), (int, float)),)
     * ...

    Raises
    ------
    ValueError
        If the type description is not valid or not recognized.

    Note
    ----
        The type description is lowercased and spaces are removed before parsing.
    """
    # Clean up
    old_type_desc = type_desc
    type_desc = type_desc.replace(" ", "").lower()
    try:
        # Base typeParseType
        base_type = _parse_base_type(type_desc)
        if base_type is not None:
            return (base_type,)
        # Split the external blocks enclosed by brackets
        blocks = _split_brackets(type_desc, delimiter="|")
        types: Tuple = ()
        for block in blocks:
            if block[:5] == "list[":
                types += _parse_list(block)
            elif block[:5] == "dict[":
                types += _parse_dict(block)
            elif block[:9] == "optional[":
                types += _parse_optional(block)
            elif block[:6] == "union[":
                types += _parse_union(block)
            else:  # Should be a base type
                base_type = _parse_base_type(block)
                if base_type is not None:
                    types += (base_type,)
                else:
                    raise ValueError(f"Unknown type: '{block}'")
        return types
    except ValueError as err:
        # Revert the traceback to show the original type description
        raise ValueError(f"Unknown type: '{old_type_desc}'") from err


def _parse_base_type(type_desc: str) -> Optional[Type]:
    """Parse a base type description.

    Base types are: none, any, bool, int, float and str, list, dict.
    Return None if the type is not a base type.
    """
    if type_desc == "none":
        return type(None)
    if type_desc == "any":
        # Match any type
        return object
    if type_desc in ("bool", "int", "float", "str", "list", "dict"):
        return locate(type_desc)  # type: ignore
    return None


def _parse_list(type_desc: str) -> Tuple:
    """Parse a "list" type description."""
    sub_desc = type_desc[5:-1]
    if len(_split_brackets(sub_desc, delimiter=",")) > 1:
        raise ValueError(f"Invalid List type: '{type_desc}'")
    return (("list",) + (_parse_type(sub_desc),),)


def _parse_dict(type_desc: str) -> Tuple:
    """Parse an "dict" type description."""
    sub_desc = type_desc[5:-1]
    sub_blocks = _split_brackets(sub_desc, delimiter=",")
    if len(sub_blocks) != 2:
        raise ValueError(f"Invalid Dict type: '{type_desc}'")
    key_type = _parse_type(sub_blocks[0])
    value_type = _parse_type(sub_blocks[1])
    return (("dict",) + (key_type,) + ((value_type),),)


def _parse_optional(type_desc: str) -> Tuple:
    """Parse an "optional" type description."""
    sub_desc = type_desc[9:-1]
    if len(_split_brackets(sub_desc, delimiter=",")) > 1:
        raise ValueError(f"Invalid Optional type: '{type_desc}'")
    return ((type(None),) + _parse_type(sub_desc),)


def _parse_union(type_desc: str) -> Tuple:
    """Parse an "union" type description."""
    sub_desc = type_desc[6:-1]
    # Split by comma
    sub_blocks = _split_brackets(sub_desc, delimiter=",")
    if len(sub_blocks) < 2:
        raise ValueError(f"Invalid Union type: '{type_desc}'")
    types: Tuple = ()
    union_types = [_parse_type(sub_block) for sub_block in sub_blocks]
    for union_type in union_types:
        types += union_type
    return types


def _split_brackets(type_desc: str, delimiter: str) -> List[str]:
    """Split a type description in blocks enclosed by brackets."""
    blocks = []
    bracket_count = 0
    i, j = 0, 0
    while j < len(type_desc):
        char = type_desc[j]
        if char == "[":
            bracket_count += 1
        elif char == "]":
            bracket_count -= 1
        if bracket_count == 0 and char == delimiter:
            blocks.append(type_desc[i:j])
            i = j + 1
            j += 2
        else:
            j += 1
    blocks.append(type_desc[i:j])
    return blocks


def _isinstance(obj: object, types: Union[Type, Tuple]) -> bool:
    """Check if an object is an instance of a type or a tuple of types.

    Intended to work with the outputs of _parse_type.
    """
    if isinstance(types, type):
        return isinstance(obj, types)
    if types[0] == "list" and len(types) == 2:
        return isinstance(obj, list) and all(
            _isinstance(elem, types[1]) for elem in obj
        )
    if types[0] == "dict" and len(types) == 3:
        return (
            isinstance(obj, dict)
            and all(_isinstance(key, types[1]) for key in obj)
            and all(_isinstance(value, types[2]) for value in obj.values())
        )
    if isinstance(types[0], (type, tuple)):
        return any(_isinstance(obj, sub_types) for sub_types in types)
    raise ValueError(f"Invalid type for _isinstance: '{types}'")


def _convert_type(obj: object, types: Union[Type, Tuple]) -> object:
    """Try to convert an object to a type or a tuple of types.

    Intended to work with the outputs of _parse_type.
    """
    try:
        return _convert_type_internal(obj, types)
    except (TypeError, ValueError):
        return obj


def _convert_type_internal(obj: object, types: Union[Type, Tuple]) -> object:
    """Try to convert an object to a type or a tuple of types.

    Intended to work with the outputs of _parse_type.
    """
    if isinstance(types, type):
        return types(obj)
    if types[0] == "list" and len(types) == 2:
        return [_convert_type_internal(elem, types[1]) for elem in obj]  # type: ignore
    if types[0] == "dict" and len(types) == 3:
        return {
            _convert_type_internal(key, types[1]): _convert_type_internal(
                value, types[2]
            )
            for key, value in obj.items()  # type: ignore
        }
    if isinstance(types[0], (type, tuple)):
        if any(_isinstance(obj, sub_types) for sub_types in types):
            return obj
        for sub_types in types:
            try:
                return _convert_type_internal(obj, sub_types)
            except (TypeError, ValueError):
                pass
    raise TypeError



# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Integration test for sweep updates."""
import pytest_check as check

from cliconfig import make_config, update_config


def test_sweep() -> None:
    """Integration test for sweep."""
    config = make_config("tests/configs/integration/sweep.yaml", no_cli=True)
    expected_dict = {
        "args": {"a": 5, "b": 3, "c": True, "d": [1, 0]},
        "res": [1, 0, 3, 6],
    }
    check.equal(config.dict, expected_dict)

    config = update_config(
        config, {"args": {"a": 2, "b": 7, "c": False, "d": [1, 2, 3]}}
    )
    expected_dict = {
        "args": {"a": 2, "b": 7, "c": False, "d": [1, 2, 3]},
        "res": [1, 2, 3, 0, 7],
    }
    check.equal(config.dict, expected_dict)
    config = update_config(config, {"args": {"a": 2}, "res": 5})
    check.equal(config.res, 5)
    config = update_config(config, {"args": {"a": 3}})
    check.equal(config.res, 5)



# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Test the AST parser module."""
import ast
import re

import numpy as np
import pytest
import pytest_check as check

from cliconfig.processing._ast_parser import _process_node


def test_ast_parser() -> None:
    """Test the AST parser module."""
    flat_dict: dict = {"cfg1.cfg2.cfg3.param1": 4.2, "param2": 6.2}
    expr = "(1 + 2 * cfg1.cfg2.cfg3.param1 + (param2 // 2) ** 2 - 3 / 4) % 10"
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.almost_equal(result, 7.65)

    flat_dict = {"param1": 0.2, "param2": False}
    expr = (
        "tuple([param1 if (param1 > 0.1) and (param1 <= 0.2) & "
        "(param1 == 0.2) and (param2 | True) else 0] * 2)"
    )
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.equal(result, (0.2, 0.2))

    expr = "sum([1 for _ in range(2)])"
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.equal(result, 2)

    flat_dict = {"elems": [(1, 2), (3, 4), (5, 6)], "val": 2}
    expr = "{'list': [i+2*j for i, j in elems if i > val]}, val"
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.equal(result, ({"list": [11, 17]}, 2))

    flat_dict = {"a": {"b": 1}}
    expr = "list(a.keys())"
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.equal(result, ["b"])

    flat_dict = {"elems": [1, 2, 3], "val": 5}
    expr = "np.array(elems + [random.randint(0, val)]), np.random.randint(0, 5)"
    tree = ast.parse(expr, mode="eval")
    result = _process_node(node=tree.body, flat_dict=flat_dict)
    check.is_instance(result[0], np.ndarray)
    check.is_true(0 <= result[1] < 5)
    check.is_true(np.allclose(result[0][:3], [1, 2, 3]))

    # Case not valid parameter value
    flat_dict = {"param1": 2, "param2": "string"}  # type: ignore
    tree = ast.parse("param1 + param2", mode="eval")
    with pytest.raises(
        ValueError,
        match="Invalid value in expression for parameter 'param2'.*",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    # Case unknown parameter
    tree = ast.parse("param1 + unknown", mode="eval")
    with pytest.raises(
        ValueError,
        match="Unknown parameter 'unknown'.",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    # Case invalid node
    tree = ast.parse("lambda x: 0", mode="eval")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Not supported node in expression (of type <class 'ast.Lambda'>)."
        ),
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    # Case invalid binary operator
    tree = ast.parse("param1 >> param1", mode="eval")
    with pytest.raises(
        ValueError,
        match="Invalid operator detected: <ast.RShift.*",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    # Case invalid compare operator
    tree = ast.parse("param1 >> param1", mode="eval")
    with pytest.raises(
        ValueError,
        match="Invalid operator detected: <ast.RShift.*",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    # Case invalid package or function
    tree = ast.parse("np.save('tmp.npy', np.array([1, 2, 3]))", mode="eval")
    with pytest.raises(
        ValueError,
        match="Package or function not allowed or not supported: numpy.save",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)

    tree = ast.parse("random._exp()", mode="eval")
    with pytest.raises(
        ValueError,
        match="Package or function not allowed or not supported: random._exp",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)
    tree = ast.parse("logml.Logger()", mode="eval")
    with pytest.raises(
        ValueError,
        match="Package or function not allowed or not supported: logml.Logger",
    ):
        _process_node(node=tree.body, flat_dict=flat_dict)



# Copyright (c) 2023 Valentin Goldite. All Rights Reserved.
"""Processing module to apply functions on merge (before and after), save and load."""

from cliconfig.processing.base import Processing

__all__ = ["Processing"]
